//****************************************************************************
//      Linux Host Driver for SiI SATA RAID Adapters
//
// sii_io.c - LINUX SCSI low level driver functions
//
// Copyright (c) 2005  Silicon Image.  All rights reserved.
//
// Creator: Vu Le
//
//****************************************************************************
#include <siidriver.h>

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
#define SCSI_DATA_UNKNOWN   (DMA_BIDIRECTIONAL)
#define SCSI_DATA_WRITE     (DMA_TO_DEVICE)
#define SCSI_DATA_READ      (DMA_FROM_DEVICE)
#define SCSI_DATA_NONE      (DMA_NONE)

#define scsi_to_pci_dma_dir(scsi_dir)   ((int)(scsi_dir))
#define scsi_to_sbus_dma_dir(scsi_dir)  ((int)(scsi_dir))
#endif

static int
isEmptyChanIO(sii_controller_t *contrl, int chan)
{
    return (!contrl->dev_chan_queue[chan].io_chan_count);
}

static int
isEmptyAlChanIO(sii_controller_t *contrl)
{
    int i;
    int rlt = 0;

    if (!contrl)
	return (rlt);

    for (i = 0; i < contrl->CHANNELS_PER_CONTROLLER; i++) {
	rlt += isEmptyChanIO(contrl, i);
    }
    return (rlt);
}

static void*
sii_get_srb(sii_controller_t *contrl)
{
    PSCSI_REQUEST_BLOCK srb = NULL;
                                                                                
    if (contrl) {
        if (contrl->dev_free_srb_list == NULL) {
            // Allocate and Add to the free list
	    sii_mutex_exit(&contrl->dev_io_mtx);
            srb = sii_srb_alloc(contrl);
    	    sii_mutex_enter(&contrl->dev_io_mtx);
        } else {
            // Remove from the free list
            srb = contrl->dev_free_srb_list;
            contrl->dev_free_srb_list =
                contrl->dev_free_srb_list->NextSrb;
            srb->NextSrb = NULL;
        }
        return (srb);
    } else
        return (NULL);
}

static void
sii_toss_srb(DPSCSI_REQUEST_BLOCK srb_hd, sii_controller_t *contrl)
{
    PSCSI_REQUEST_BLOCK srb = NULL;
    PSCSI_REQUEST_BLOCK curr = NULL;
                                                                                
    if (contrl && srb_hd) {
        srb = *srb_hd;
        *srb_hd = NULL;
                                                                                
        curr = contrl->dev_free_srb_list;
        contrl->dev_free_srb_list = srb;
        srb->NextSrb = curr;
    }
}



//****************************************************************
//	CreateSRB
//	This routine will create a SCSI_REQUEST_BLOCK based on original
//	scsi_cmnd structure
//
//	Arguments:
//		SCpnt - Pointer to SCSI command structure.
//		pSRB  - Pointer to SRB structure
// Returns:
//		TRUE if all goes well.
//
//****************************************************************
static sii_u32
CreateSRB(struct scsi_cmnd *SCpnt, SCSI_REQUEST_BLOCK **pSRB)
{
    PSCSI_REQUEST_BLOCK	Srb = NULL;
    sii_controller_t    *contrl = NULL;

    if (!SCpnt)
	return (FALSE);
    contrl = HOSTDATA(SCpnt->device->host);

    Srb = sii_get_srb(contrl);
    if(!Srb) {
	//ERR_MSG("CreateSRB:MALLOC FAILED\n");
    sii_printd("CreateSRB:MALLOC FAILED\n");
	return (FALSE);
    }

    memset(SCpnt->sense_buffer, 0x00, SCSI_SENSE_BUFFERSIZE);
    memcpy(Srb->Cdb, SCpnt->cmnd, SCpnt->cmd_len);

    Srb->Function	= SRB_FUNCTION_EXECUTE_SCSI;
    Srb->OriginalRequest	= SCpnt;
    Srb->PathId		= SCpnt->device->channel;
    Srb->TargetId	= SCpnt->device->id;
    Srb->Lun		= SCpnt->device->lun;
    Srb->CdbLength	= SCpnt->cmd_len;
    Srb->use_sg		= SCpnt->use_sg;

    if(SCpnt->sc_data_direction == SCSI_DATA_READ)
	Srb->SrbFlags	= SRB_FLAGS_DATA_IN;
    else if(SCpnt->sc_data_direction == SCSI_DATA_WRITE)
	Srb->SrbFlags	= SRB_FLAGS_DATA_OUT;
    else
	Srb->SrbFlags	= 0;

    Srb->SrbStatus		= SRB_STATUS_PENDING;
    Srb->DataTransferLength	= SCpnt->request_bufflen;
    Srb->DataBuffer		= SCpnt->request_buffer;
    Srb->SenseInfoBuffer	= SCpnt->sense_buffer;
    Srb->SenseInfoBufferLength	= SCSI_SENSE_BUFFERSIZE;
    SCpnt->host_scribble	= (PUCHAR) Srb;

    Srb->Physical_Addr = Srb->Physical_Addr2 = (SCSI_PHYSICAL_ADDRESS)NULL;  

    // XXX vuale cleanup: need to map to pci memory here
    if (SCpnt->use_sg){
    	if (SCpnt->use_sg > MAX_SG_TABLE_ENTRIES) {
            //ERR_MSG("SiiDriver Scatter Gather Table Overflow!")
            sii_printd("SiiDriver Scatter Gather Table Overflow!\n");
            return (FALSE);
        }
#ifdef DEBUG_SII_SCSI_HAS_HOST_LOCK
	sii_u32* p;
	sii_u32  i;
        p = (sii_u32 *) SCpnt->request_buffer;
        for (i = 0; i < SCpnt->use_sg ; i++) {
              MSG("%x ",*p++)
              MSG("\n")
        }
#endif
	SiI_pci_map_sg((PVOID)contrl->dev_pcidev,
		(PVOID) SCpnt->request_buffer, SCpnt->use_sg,
		SiI_scsi_to_pci_dma_dir(Srb->SrbFlags));

    } else {
/*
        Srb->Physical_Addr = SiI_pci_map_page(contrl->dev_pcidev,
                           	Srb->DataBuffer,
				(Srb->DataTransferLength + 0x10000),
				SiI_scsi_to_pci_dma_dir(Srb->SrbFlags));
*/
    }
    Srb->OrigSRB = (PSCSI_REQUEST_BLOCK)(-1);  // MARK SRB as mapped
    *pSRB = Srb;
    return (TRUE);
}

static void
ConvertResult(PSCSI_REQUEST_BLOCK Srb)
{
    sii_u32             Result;
    sii_u32             Status;


    if (!Srb)
	return;
    Status  = SRB_STATUS(Srb->SrbStatus);
                                                                                
    switch(Status) {
    case SRB_STATUS_SUCCESS:
    	Result  = DID_OK << 16;
        break;
    case SRB_STATUS_ERROR:
        DEBUG("SiI: DID_ERROR\n");
        if( Srb->SrbStatus & SRB_STATUS_AUTOSENSE_VALID) {
            Result  =
		(DRIVER_SENSE<<24) | (DID_ERROR << 16) | (CHECK_CONDITION<<1);
 	} else {
            Result =
		(SUGGEST_SENSE<<24) | (DID_ERROR << 16) | (CHECK_CONDITION<<1);
	}
        break;
     case SRB_STATUS_SELECTION_TIMEOUT:
     case SRB_STATUS_NO_DEVICE:
        Result = DID_BAD_TARGET << 16;
        break;
     case    SRB_STATUS_TIMEOUT:
        DEBUG("SiI, timeout - channel=%d report as bad target\n",
								Srb->PathId);
        Result = DID_TIME_OUT << 16;
        break;
     case    SRB_STATUS_ABORTED:
        DEBUG("SiI, Aborted channel=%d abort\n",Srb->PathId);
        Result = DID_ABORT << 16;
     	break;
    case    SRB_STATUS_BUS_RESET:
   	    DEBUG("SiI, reset channel=%d reset\n",Srb->PathId);
        Result = DID_RESET << 16;
        break;
    case    SRB_STATUS_PARITY_ERROR:
        DEBUG("SiI parity error channel=%d parity\n",Srb->PathId);
        Result = DID_PARITY << 16;
        break;
    case    SRB_STATUS_INVALID_PATH_ID:
    case    SRB_STATUS_INVALID_TARGET_ID:
    case    SRB_STATUS_INVALID_LUN:
        DEBUG("SiI, DID_BAD_TARGET::\n",Srb->PathId);
        Result = DID_BAD_TARGET << 16;
        break;
    case    SRB_STATUS_BUSY:
        Result = DID_BUS_BUSY << 16;
        break;
    default:
        Result = DID_OK<<16;
        break;
   }
   ((struct scsi_cmnd *)(Srb->OriginalRequest))->result = Result;
}

/*
 * Always protected by io_queue_lock
 * 	Got here from Interrupt()
 * The key function to release an IO
 */
static void
LSP_Done(sii_controller_t *contrl, SCSI_REQUEST_BLOCK *Srb)
{
    struct scsi_cmnd	*pScmnd;

    if (!contrl || !Srb)
	return;

    pScmnd = (struct scsi_cmnd *) Srb->OriginalRequest;
    if (pScmnd != NULL) {
    	// Unmap ONCE
    	UnMapSGBuffer(contrl,Srb);

    	/*
     	 * convert result to linux representation
     	 */  
    	ConvertResult(Srb);
    	pScmnd->host_scribble = 0;
    	sii_toss_srb(&Srb, contrl);
    	pScmnd->scsi_done(pScmnd);
    } else if (Srb->Internal_cmd != NULL) {
	up(&contrl->dev_ioctl_sem);
    }
}

/*
 * Always protected by io_queue_lock
 * 	Got here from Interrupt()
 */

asmlinkage void
LSPNotify(IN SCSI_NOTIFICATION_TYPE NType, IN PVOID DeviceExt, PVOID p0, ULONG p1,...)
{
    PSCSI_REQUEST_BLOCK Srb = NULL;
    sii_controller_t	*contrl = NULL;
    sii_u16		rev_dev_map = 0;
    sii_u16		add_dev_map = 0;

    if (!DeviceExt) {
	sii_print("Got Null DeviceExt passed in LSPNotify()\n");
	return;
    }
    contrl = outer(sii_controller_t, _hw_dev_ext_, DeviceExt);

    switch (NType) {
    case RequestComplete:
	Srb     = (PSCSI_REQUEST_BLOCK)p0;
    	LSP_Done(contrl, Srb);
	break;

    case RequestTimerCall:
	// Try to delete the timer
        if (p1 == 0) {
            SiI_del_timer(&contrl->siiTimer);
            sii_disable_raid_timer(DeviceExt);
            break;
        }

	sii_RequestTimerCallback(DeviceExt, p1, (PVOID_CB)p0);
	break;
    case BusChangeDetected:
    	rev_dev_map = (p1 >> 16);
	add_dev_map = (p1 & 0xffff);
	// Stop all I/O on to this device
	if (rev_dev_map) {
	    contrl->dev_chan_queue[(sii_ulong)p0].io_chan_count = NOBUS;
	    contrl->dev_active_dmap[(sii_ulong)p0] &= ~rev_dev_map;
	}

	if (add_dev_map) {
	    contrl->dev_chan_queue[(sii_ulong)p0].io_chan_count = 0;
	    contrl->dev_active_dmap[(sii_ulong)p0] |= add_dev_map;
	}

	contrl->dev_change_dmap[(sii_ulong)p0] |= p1;
	contrl->dev_flag |= SII_HOTPLUG;
        up(&contrl->dev_io_sem);
	break;
    default:
	// XXX vuale
	// sii_print(
	//	"Not handle this for LSPNotify() yet %d\n", NType);
	break;
    }
}

asmlinkage ULONG
LSPSetBusDataBO(
	IN PVOID	DeviceExt, 		IN ULONG	BusDataType,
	IN ULONG	SystemIOBusNumber, 	IN ULONG	SlotNumber,
	IN PVOID	Buffer, 		IN ULONG	Offset,
	IN ULONG	Length)
{
    sii_controller_t	*contrl = NULL;
    int 		i	= 0;
    sii_u8		*byte 	= NULL;
    if (!DeviceExt) {
	sii_print("Got Null DeviceExt passed in LSPNotify()\n");
	return (0);
    }
    contrl = outer(sii_controller_t, _hw_dev_ext_, DeviceExt);

    switch (BusDataType) {
    case PCIConfiguration:
	byte = (sii_u8*)Buffer;
    	for(i = 0; i < Length; i++)
            if (SiI_pci_write_config_byte(
			(PVOID)contrl->dev_pcidev,
			(sii_u8)(Offset + i), byte[i]))
	    	break;
	break;
    default:
	i = 0;
	break;
    }

    return (i);
}

asmlinkage ULONG
LSPGetBusData(
	IN PVOID	DeviceExt, 		IN ULONG	BusDataType,
	IN ULONG	SystemIOBusNumber, 	IN ULONG	SlotNumber,
	IN PCHAR	Buffer, 		IN ULONG	Length)
{
    sii_controller_t	*contrl = NULL;
    int 		i = 0;

    if (!DeviceExt) {
	sii_print("Got Null DeviceExt passed in LSPNotify()\n");
	return (0);
    }
    contrl = outer(sii_controller_t, _hw_dev_ext_, DeviceExt);

    switch (BusDataType) {
    case PCIConfiguration:
    	for(i = 0; i < Length; i++) {
            if (SiI_pci_read_config_byte(
			(PVOID)contrl->dev_pcidev, i, &Buffer[0])) {
	    	break;
	    }
	}
	break;
    default:
	i = 0;
	break;
    }

    sii_dev_ext_init((PVOID)&contrl->_hw_dev_ext_[0],
      contrl->dev_pcidev->irq,
      contrl->dev_pcidev->bus->number,
      contrl->dev_pcidev->devfn);

    if (i == 256) {
       i = sizeof (PCI_COMMON_CONFIG);
    }

    return (i);
}

/*
 * This always is protected by dev_io_mtx
 */

static void
sii_io_send(sii_controller_t *contrl, struct scsi_cmnd *SCpnt)
{
    SCSI_REQUEST_BLOCK	*pSRB = NULL;
    //PVOID	DeviceExt = NULL;
    sii_u32		ret;

    if (!contrl || !SCpnt)
	return;

    if(CreateSRB(SCpnt, &pSRB) == FALSE) {
	MSG("Error Creating SRB\n");
	SCpnt->result = DID_BAD_TARGET << 16; // Done
	SCpnt->scsi_done (SCpnt);
	return;
    }

//    sii_printDdb(" --- sii_io_send ", pSRB);
    LSPApi(1, (void **)contrl, pSRB, 0);
    ret = pSRB->SrbStatus;

    if(ret == SRB_STATUS_BUSY) {
	LSP_Done(contrl, pSRB);
    }
}

//****************************************************************
//	sii_QueueCommand
//	Process a queued command from the SCSI manager.
//
//	Arguments:
//		SCpnt - Pointer to SCSI command structure.
//		done  - Pointer to done function to call.
//	Returns:
//		Status code.
// ****************************************************************/
int
sii_QueueCommand(struct scsi_cmnd *SCpnt, void (*done)(struct scsi_cmnd *))

{
    sii_controller_t	*contrl = HOSTDATA(SCpnt->device->host);
    sii_request_queue_t *srq = NULL;
    sii_u8              *cmd = NULL;

    if (SCpnt->retries > 0) {
//#if 1
#ifdef DWG_DEBUG
        sii_printd("retry command SiI%d, channel=%d\n",
                        SCpnt->device->host->host_no, SCpnt->device->channel);
        SiI_scsi_print_command(SCpnt);
#endif // DWG_DEBUG
    }

    if (done == NULL) {
	sii_printd("\r\n --- sii_queuecommand: %02X: done can't be NULL\n",
		SCpnt->cmnd[0]);
	SCpnt->result = DID_BAD_TARGET << 16; // Done
	return (DID_BAD_TARGET);
    }

    SCpnt->result = 0;
    SCpnt->scsi_done = done;
    SCpnt->host_scribble = 0;

    /* Do not send the command down to a non-existent device */
//    sii_printd(" --- sii_Q_Cmd channel: %x, id %x\n", SCpnt->device->channel, SCpnt->device->id);
    //if (0)//!(contrl->dev_active_dmap[SCpnt->device->channel] & (1 << SCpnt->device->id)))
    if (!(contrl->dev_active_dmap[SCpnt->device->channel] & (1 << SCpnt->device->id)))
    {
        sii_printd("\r\n ---sii_queuecommand: device channel:%d target:%d is not present (%08X)\n", 
                   SCpnt->device->channel, SCpnt->device->id, contrl->dev_active_dmap[SCpnt->device->channel]);
        SCpnt->result = (DID_BAD_TARGET << 16);
        SCpnt->scsi_done(SCpnt);
        return 0;
    }

    srq = (sii_request_queue_t*)&(SCpnt->SCp);
    srq->srq_oricmd = SCpnt;
    cmd = (sii_u8*)SCpnt->cmnd;

    sii_mutex_enter(&contrl->dev_io_mtx);
    switch (cmd[0]) {
    default:
	sii_io_send(contrl, SCpnt);
	break;
    }
    sii_mutex_exit(&contrl->dev_io_mtx);
    return (0);
}

static int
sii_io_complete_loop(void *arg)
{
    sii_controller_t *contrl = (sii_controller_t *)arg;
    sii_u32		local_devmap = 0;
    sii_u16		add_dev_map = 0;
    sii_u16		rev_dev_map = 0;
    sii_u32		i, j;

    /* Become our own separate kernel thread */
    lock_kernel();
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,60))
    daemonize("sii_io_%d", contrl->ControllerId);
#else
    /* Don't care about any signals */
    siginitsetinv(&current->blocked, 0);

    daemonize();
    sprintf(current->comm, "sii_io_%d", contrl->ControllerId);
#endif
    unlock_kernel();

    for (;;) {
        sii_ulong   dev_flag;

        (void) down_interruptible(&contrl->dev_io_sem);

        sii_mutex_enter(&contrl->dev_io_mtx);
        dev_flag = contrl->dev_flag;
        contrl->dev_flag &= ~dev_flag;
        sii_mutex_exit(&contrl->dev_io_mtx);

	if (dev_flag & SII_END) {
            if(!(dev_flag & ~(SII_END)) &&
		(isEmptyAlChanIO(contrl))) { // not doing anything else
            	contrl->dev_flag = SII_INIT;

                complete_and_exit(&contrl->dev_io_loop_exited, 0);
            	return (TRUE);
	    } else {
                // Schedule ourselves to run again
                up(&contrl->dev_io_sem);
	    }
	}

	if (dev_flag & SII_HOTPLUG) {
	    for (j = 0; j < contrl->CHANNELS_PER_CONTROLLER; j++) {
	    	rev_dev_map = (contrl->dev_change_dmap[j] >> 16);
	    	add_dev_map = (contrl->dev_change_dmap[j] & 0xffff);
	    	local_devmap = contrl->dev_change_dmap[j];

	    	if (rev_dev_map != 0) {
	    	    for(i = 0; i < contrl->DEVICES_PER_CHANNEL; i++) {
		    	if ((rev_dev_map >> i) & 0x0001) {
	    	    	    sii_scsi_scan(contrl, j, i, FALSE);
		    	}
	    	    }
	    	}

		if (add_dev_map != 0) {
	    	    for (i = 0; i < contrl->DEVICES_PER_CHANNEL; i++) {
		    	if ((add_dev_map >> i) & 0x0001) {
	    	    	    sii_scsi_scan(contrl, j, i, TRUE);
		    	}
	    	    }
	      	}
		
		// No new change since the time droping the lock
		if (contrl->dev_change_dmap[j] == local_devmap) {
	    	    contrl->dev_change_dmap[j] = 0;
		}
	    }
        }
    }
                                                                                
    sii_print("If we are here, we are in deep sh@#!!!\n");
    complete_and_exit(&contrl->dev_io_loop_exited, 0);
    return (TRUE);
}


void
sii_io_complete_thread_init(sii_controller_t *contrl)
{
    sii_u32 thid = 0;
                                                                                
    thid = sii_thread_create(sii_io_complete_loop, (void*)contrl);
    if (!thid) {
        sii_print("Can't create io complete thread for controller %s\n",
                DEVICE_NAME);
    }
}

